# Copyright © 2021-2022 Arrikto Inc.  All Rights Reserved.

"""Kale SDK.

This script creates a PyTorch distributed KFP step to solve Fashion MNIST.
"""

import time
import torch
import torch.nn as nn
import torchvision.models as models

from torchvision import datasets
from torchvision.transforms import ToTensor
from torch.utils.data import DataLoader
from rok_kubernetes.exceptions import ApiException

from kale.sdk import pipeline, step
from kale.distributed import pytorch


def _train_step(model, data_loader, criterion, optimizer, device, args):
    import logging

    log = logging.getLogger(__name__)

    for i, (features, labels) in enumerate(data_loader):
        features = features.to(device)
        labels = labels.to(device)

        pred = model(features)
        loss = criterion(pred, labels.reshape(-1))

        loss.backward()

        optimizer.step()
        optimizer.zero_grad()

        if i % args.get("log_interval", 2) == 0:
            log.info("\tLoss = {}".format(loss.item()))

    return loss


def _eval_step(model, data_loader, criterion, device, args):
    import logging

    log = logging.getLogger(__name__)

    for i, (features, labels) in enumerate(data_loader):
        features = features.to(device)
        labels = labels.to(device)

        pred = model(features)
        loss = criterion(pred, labels.reshape(-1))

    log.info("\tLoss = {}".format(loss.item()))

    return loss


def _load_data():
    """Load and prepare the data."""
    train_data = datasets.FashionMNIST(
        root="/home/jovyan/data",
        train=True,
        download=True,
        transform=ToTensor()
    )

    valid_data = datasets.FashionMNIST(
        root="/home/jovyan/data",
        train=False,
        download=True,
        transform=ToTensor()
    )

    train_data_loader = DataLoader(train_data, batch_size=1024, shuffle=True)
    valid_data_loader = DataLoader(valid_data, batch_size=2014)

    return train_data_loader, valid_data_loader


@step(name="create_model")
def create_model():
    """Define the model."""
    model = models.resnet18(pretrained=True)
    model.conv1 = nn.Conv2d(1, 64, kernel_size=(7, 7), stride=(2, 2),
                            padding=(3, 3), bias=False)

    for param in model.parameters():
        param.requires_grad = False

    features_in = model.fc.in_features
    model.fc = nn.Linear(features_in, 10)

    optimizer = torch.optim.Adam(model.parameters(), lr=.03)
    criterion = nn.CrossEntropyLoss()

    return model, optimizer, criterion


@step(name="model_training")
def distribute_training(model, optimizer, criterion):
    """Train the model."""
    train_data_loader, valid_data_loader = _load_data()
    job = pytorch.distribute(model,
                             train_data_loader,
                             criterion,
                             optimizer,
                             _train_step,
                             eval_data_loader=valid_data_loader,
                             eval_step=_eval_step,
                             epochs=1,
                             number_of_processes=2,
                             cuda=True,
                             train_args={"log_interval": 2})

    return job.name


@step(name="monitor")
def monitor(name):
    """Monitor the PyTorchJob CR."""
    job = pytorch.PyTorchJob(name)
    while True:  # Iterate if streaming logs raises an ApiException
        while True:
            cr = job.get()
            if (job.get_job_status() not in ["", "Created", "Restarting"]
                    and (cr.get("status", {})
                           .get("replicaStatuses", {})
                           .get("Master", {}))):
                break
            print("Job pending...")
            time.sleep(2)
        try:
            job.stream_logs()
            break
        except ApiException as e:
            print("Streaming the logs failed with: %s" % str(e))
            print("Retrying...")


@step(name="delete")
def delete(name):
    """Delete the PyTorchJob CR."""
    job = pytorch.PyTorchJob(name)
    while job.get_job_status() != "Succeeded":
        continue
    job.delete()


@pipeline(name="fmnist", experiment="kale-dist")
def ml_pipeline():
    """Run the ML pipeline."""
    model, optimizer, criterion = create_model()
    name = distribute_training(model, optimizer, criterion)
    monitor(name)
    delete(name)


if __name__ == "__main__":
    ml_pipeline()
